/*
 *		This Code Was Created By Jeff Molofee 2000
 *		A HUGE Thanks To Fredric Echols For Cleaning Up
 *		And Optimizing The Base Code, Making It More Flexible!
 *		If You've Found This Code Useful, Please Let Me Know.
 *		Visit My Site At nehe.gamedev.net
 */

#include <windows.h>		// Header File For Windows
#include <stdio.h>			// Header File For Standard Input/Output
#include <stdarg.h>			// Header File For Variable Argument Routines
#include <math.h>			// Header File For Math Functions
#include <time.h>			// Header File For time() Function
#include <gl\gl.h>			// Header File For The OpenGL32 Library
#include <gl\glu.h>			// Header File For The GLu32 Library
#include <gl\glaux.h>		// Header File For The Glaux Library

HDC			hDC=NULL;		// Private GDI Device Context
HGLRC		hRC=NULL;		// Permanent Rendering Context
HWND		hWnd=NULL;		// Holds Our Window Handle
HINSTANCE	hInstance;		// Holds The Instance Of The Application

GLuint		base;			// Base Display List For The Font Set

bool	keys[256];			// Array Used For The Keyboard Routine
bool	active=TRUE;		// Window Active Flag Set To TRUE By Default
bool	fullscreen=TRUE;	// Fullscreen Flag Set To Fullscreen Mode By Default

#define TITLEBAR	"Mark Morley's OpenGL Frustum Culling Tutorial"	// Title For Window
#define MAXOBJECTS	1000	// Max Number Of Objects (Points, Spheres, and Boxes Combined) To Create

int		numobjects = 200;	// Holds The Number Of Objects To Draw
int		windowwidth;		// Holds The Current Window (Screen) Width
int		windowheight;		// Holds The Current Window (Screen) Height
long	frame = 0;			// Counter To Be Used In Calculating FPS
long	timebase = 0;		// Used In Calculating FPS
float	fps = 0.0f;			// Holds The Current FPS (Frames Per Second)
bool	culling = false;	// Indicates If Frustum Culling Is On Or Not
bool	grid = true;		// Indicates If Grid Should Be Drawn
int		mode = 1;			// Indicates Object Mode: 0=points 1=spheres 2=cubes
float	frustum[6][4];		// Holds The Current Frustum Plane Equations
float	fov = 45.0f;		// Holds The Current FOV Value
float	xpos = 0.0f;		// Holds The Camera's Current X Position
float	zpos = 0.0f;		// Holds The Camera's Current Z Position
float	tilt = 0.0f;		// Holds The Camera's Current Tilt Angle
float	heading = 0.0f;		// Holds The Camera's Current Heading
float	objects[MAXOBJECTS][7];	// Holds The Center X, Y, Z, Radius, and RGB For MAXOBJECTS Objects
GLfloat LightAmbient[]= { 0.2f, 0.2f, 0.2f, 1.0f };		// Ambient Light Values
GLfloat LightDiffuse[]= { 1.0f, 1.0f, 1.0f, 1.0f };		// Diffuse Light Values
GLfloat LightPosition[]= { 0.0f, 0.0f, 0.0f, 1.0f };	// Light Is At Origin

LRESULT	CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);	// Declaration For WndProc

GLvoid ReSizeGLScene(GLsizei width, GLsizei height)				// Resize And Initialize The GL Window
{
	if (height==0)										// Prevent A Divide By Zero By
	{
		height=1;										// Making Height Equal One
	}

	glViewport(0,0,width,height);						// Reset The Current Viewport

	glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
	glLoadIdentity();									// Reset The Projection Matrix

	// Calculate The Aspect Ratio Of The Window
	gluPerspective(fov,(GLfloat)width/(GLfloat)height,0.5f,50.0f);

	glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	glLoadIdentity();									// Reset The Modelview Matrix

	windowwidth = width;								// Save Width In Global
	windowheight = height;								// Save Height In Global
}

GLvoid BuildFont(GLvoid)								// Build Our Bitmap Font
{
	HFONT	font;										// Windows Font ID

	base = glGenLists(96);								// Storage For 96 Characters
	font = CreateFont(	-12,							// Height Of Font
						0,								// Width Of Font
						0,								// Angle Of Escapement
						0,								// Orientation Angle
						FW_NORMAL,						// Font Weight
						FALSE,							// Italic
						FALSE,							// Underline
						FALSE,							// Strikeout
						ANSI_CHARSET,					// Character Set Identifier
						OUT_TT_PRECIS,					// Output Precision
						CLIP_DEFAULT_PRECIS,			// Clipping Precision
						ANTIALIASED_QUALITY,			// Output Quality
						FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
						"Courier New");					// Font Name
	SelectObject(hDC, font);							// Selects The Font We Created
	wglUseFontBitmaps(hDC, 32, 96, base);				// Builds 96 Characters Starting At Character 32
}

GLvoid KillFont(GLvoid)									// Delete The Font
{
 	glDeleteLists(base, 96);							// Delete All 96 Characters
}

GLvoid glPrint(float x, float y, const char *fmt, ...)	// Custom GL "Print" Routine
{
	char		text[256];								// Holds Our String
	va_list		ap;										// Pointer To List Of Arguments

	if (fmt == NULL)									// If There's No Text
		return;											// Do Nothing
	va_start(ap, fmt);									// Parses The String For Variables
    vsprintf(text, fmt, ap);							// And Converts Symbols To Actual Numbers
	va_end(ap);											// Results Are Stored In Text
	glPushAttrib(GL_LIST_BIT);							// Pushes The Display List Bits
	glListBase(base - 32);								// Sets The Base Character to 32
	glRasterPos2f( x, y );								// Move To Drawing Position
	glCallLists(strlen(text), GL_UNSIGNED_BYTE, text);	// Draws The Display List Text
	glPopAttrib();										// Pops The Display List Bits
}

int InitGL(GLvoid)										// All Setup For OpenGL Goes Here
{
	int i;

	srand( (unsigned)time( NULL ) );					// Seed The Random Number Generator
	for( i = 0; i < MAXOBJECTS; i++ )					// Create MAXOBJECTS Random Objects
	{
		objects[i][0] = -100.0f + rand() % 200;			// Random X From -100 To 100
		objects[i][1] = -10.0f + rand() % 20;			// Random Y From -10 To 10
		objects[i][2] = -100.0f + rand() % 200;			// Random Z From -100 To 100
		objects[i][3] = 0.5f * (1 + rand() % 5);		// Random Size/Radius From 0.5 to 3
		objects[i][4] = .01f * (rand() % 100);			// Random RED Value
		objects[i][5] = .01f * (rand() % 100);			// Random GREEN Value
		objects[i][6] = .01f * (rand() % 100);			// Random BLUE Value
	}
	BuildFont();										// Build A Bitmap Font
	glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);				// Black Background
	glClearDepth(1.0f);									// Depth Buffer Setup
	glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
	glDepthFunc(GL_LEQUAL);								// The Type Of Depth Testing To Do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations
	glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);		// Set Up Ambient Lighting
	glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);		// Set Up Diffuse Lighting
	glEnable(GL_LIGHT1);								// Enable Light #1
	glEnable(GL_LIGHTING);								// Turn On Lighting
	glEnable(GL_COLOR_MATERIAL);						// Enable Tracking The Color Parameter
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);	// Make Ambient And Diffuse Track Color
	return TRUE;										// Initialization Went OK
}

void ortho( void )										// Changes Projection To Orthgonal View For Text
{
	glDisable(GL_DEPTH_TEST);							// Disable Depth Testing (All Text On Top!)
	glDisable(GL_LIGHTING);								// Disable Lighting
	glMatrixMode(GL_PROJECTION);						// Save The Current PROJECTION Matrix
	glPushMatrix();
	glLoadIdentity();
	glOrtho( 0, windowwidth, windowheight, 0, -1, 1 );	// Enter Orthographic Mode
	glMatrixMode(GL_MODELVIEW);							// Save The MODELVIEW Matrix
	glPushMatrix();
	glLoadIdentity();
}

void perspective( void )								// Changes Projection To Perspective View
{
	glMatrixMode( GL_PROJECTION );						// Restore The PROJECTION Matrix
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );						// Restore The MODELVIEW Matrix
	glPopMatrix();
	glEnable( GL_DEPTH_TEST );							// Enable Depth Testing
	glEnable(GL_LIGHTING);								// Enable Lighting
}

void ExtractFrustum()									// Extracts The Current View Frustum Plane Equations
{
	float	proj[16];									// For Grabbing The PROJECTION Matrix
	float	modl[16];									// For Grabbing The MODELVIEW Matrix
	float	clip[16];									// Result Of Concatenating PROJECTION and MODELVIEW
	float	t;											// Temporary Work Variable
	
	glGetFloatv( GL_PROJECTION_MATRIX, proj );			// Grab The Current PROJECTION Matrix
	glGetFloatv( GL_MODELVIEW_MATRIX, modl );			// Grab The Current MODELVIEW Matrix

	// Concatenate (Multiply) The Two Matricies
	clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
	clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
	clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
	clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

	clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
	clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
	clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
	clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

	clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
	clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
	clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
	clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

	clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
	clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
	clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
	clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];


	// Extract the RIGHT clipping plane
	frustum[0][0] = clip[ 3] - clip[ 0];
	frustum[0][1] = clip[ 7] - clip[ 4];
	frustum[0][2] = clip[11] - clip[ 8];
	frustum[0][3] = clip[15] - clip[12];

	// Normalize it
	t = (float) sqrt( frustum[0][0] * frustum[0][0] + frustum[0][1] * frustum[0][1] + frustum[0][2] * frustum[0][2] );
	frustum[0][0] /= t;
	frustum[0][1] /= t;
	frustum[0][2] /= t;
	frustum[0][3] /= t;


	// Extract the LEFT clipping plane
	frustum[1][0] = clip[ 3] + clip[ 0];
	frustum[1][1] = clip[ 7] + clip[ 4];
	frustum[1][2] = clip[11] + clip[ 8];
	frustum[1][3] = clip[15] + clip[12];

	// Normalize it
	t = (float) sqrt( frustum[1][0] * frustum[1][0] + frustum[1][1] * frustum[1][1] + frustum[1][2] * frustum[1][2] );
	frustum[1][0] /= t;
	frustum[1][1] /= t;
	frustum[1][2] /= t;
	frustum[1][3] /= t;


	// Extract the BOTTOM clipping plane
	frustum[2][0] = clip[ 3] + clip[ 1];
	frustum[2][1] = clip[ 7] + clip[ 5];
	frustum[2][2] = clip[11] + clip[ 9];
	frustum[2][3] = clip[15] + clip[13];

	// Normalize it
	t = (float) sqrt( frustum[2][0] * frustum[2][0] + frustum[2][1] * frustum[2][1] + frustum[2][2] * frustum[2][2] );
	frustum[2][0] /= t;
	frustum[2][1] /= t;
	frustum[2][2] /= t;
	frustum[2][3] /= t;


	// Extract the TOP clipping plane
	frustum[3][0] = clip[ 3] - clip[ 1];
	frustum[3][1] = clip[ 7] - clip[ 5];
	frustum[3][2] = clip[11] - clip[ 9];
	frustum[3][3] = clip[15] - clip[13];

	// Normalize it
	t = (float) sqrt( frustum[3][0] * frustum[3][0] + frustum[3][1] * frustum[3][1] + frustum[3][2] * frustum[3][2] );
	frustum[3][0] /= t;
	frustum[3][1] /= t;
	frustum[3][2] /= t;
	frustum[3][3] /= t;


	// Extract the FAR clipping plane
	frustum[4][0] = clip[ 3] - clip[ 2];
	frustum[4][1] = clip[ 7] - clip[ 6];
	frustum[4][2] = clip[11] - clip[10];
	frustum[4][3] = clip[15] - clip[14];

	// Normalize it
	t = (float) sqrt( frustum[4][0] * frustum[4][0] + frustum[4][1] * frustum[4][1] + frustum[4][2] * frustum[4][2] );
	frustum[4][0] /= t;
	frustum[4][1] /= t;
	frustum[4][2] /= t;
	frustum[4][3] /= t;


	// Extract the NEAR clipping plane.  This is last on purpose (see pointinfrustum() for reason)
	frustum[5][0] = clip[ 3] + clip[ 2];
	frustum[5][1] = clip[ 7] + clip[ 6];
	frustum[5][2] = clip[11] + clip[10];
	frustum[5][3] = clip[15] + clip[14];

	// Normalize it
	t = (float) sqrt( frustum[5][0] * frustum[5][0] + frustum[5][1] * frustum[5][1] + frustum[5][2] * frustum[5][2] );
	frustum[5][0] /= t;
	frustum[5][1] /= t;
	frustum[5][2] /= t;
	frustum[5][3] /= t;
}

// Test If A Point Is In The Frustum.
bool PointInFrustum( float x, float y, float z )
{
	int p;

	for( p = 0; p < 6; p++ )
		if( frustum[p][0] * x + frustum[p][1] * y + frustum[p][2] * z + frustum[p][3] <= 0 )
			return false;
	return true;
}

// Test If A Sphere Is In The Frustum
bool SphereInFrustum( float x, float y, float z, float radius )
{
	int p;

	for( p = 0; p < 6; p++ )
		if( frustum[p][0] * x + frustum[p][1] * y + frustum[p][2] * z + frustum[p][3] <= -radius )
			return false;
	return true;
}

// Test If A Cube Is In The Frustum
bool CubeInFrustum( float x, float y, float z, float size )
{
	int p;

	for( p = 0; p < 6; p++ )
	{
		if( frustum[p][0] * (x - size) + frustum[p][1] * (y - size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
			continue;
		if( frustum[p][0] * (x + size) + frustum[p][1] * (y - size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
			continue;
		if( frustum[p][0] * (x - size) + frustum[p][1] * (y + size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
			continue;
		if( frustum[p][0] * (x + size) + frustum[p][1] * (y + size) + frustum[p][2] * (z - size) + frustum[p][3] > 0 )
			continue;
		if( frustum[p][0] * (x - size) + frustum[p][1] * (y - size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
			continue;
		if( frustum[p][0] * (x + size) + frustum[p][1] * (y - size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
			continue;
		if( frustum[p][0] * (x - size) + frustum[p][1] * (y + size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
			continue;
		if( frustum[p][0] * (x + size) + frustum[p][1] * (y + size) + frustum[p][2] * (z + size) + frustum[p][3] > 0 )
			continue;
		return false;
	}
	return true;
}

int DrawGLScene(GLvoid)									// Here's Where We Do All The Drawing
{
	int i;
	int rendered = 0;									// Used To Count Rendered Objects
	long time;											// Used In Calculating FPS
	bool visible;										// Used To Test If An Object Was Culled

	frame++;											// Increment Frame Counter
	time = GetTickCount();								// Get Timer Value
	if( time - timebase >= 1000 )						// Has A Second Passed Since Last FPS Update?
	{
		fps = frame * 1000.0f / (time - timebase);		// Calculate New FPS
	 	timebase = time;								// Save Current Tick Count
		frame = 0;										// Start Counting Frames From Zero Again
	}

	if( keys[VK_ADD] && numobjects < MAXOBJECTS )		// Add Objects If The '+' Sign Is Pressed
		numobjects++;

	if( keys[VK_SUBTRACT] && numobjects > 1 )			// Subtract Objects If The '-' Sign Is Pressed
		numobjects--;

	if( keys[VK_UP] )									// Step Forward When UP Is Pressed
	{
		xpos -= (float) sin( heading * 3.14592/180 ) * 0.25f;
		zpos += (float) cos( heading * 3.14592/180 ) * 0.25f;
	}

	if( keys[VK_DOWN] )									// Step Backward When DOWN Is Pressed
	{
		xpos -= (float) sin( heading * 3.14592f/180 ) * -0.25f;
		zpos += (float) cos( heading * 3.14592f/180 ) * -0.25f;
	}

	if( keys[VK_LEFT] )									// Turn Left When LEFT Is Pressed
		heading -= 1;

	if( keys[VK_RIGHT] )								// Turn Right When RIGHT Is Pressed
		heading += 1;

	if( heading >= 360.0 )								// Keep Heading Between 0 and 360 Degrees (Wrap It)
		heading -= 360.0;
	else if( heading < 0.0 )
		heading += 360.0;

	if( keys['U'] )										// Tilt Head Up When 'U' Is Pressed
		tilt -= 1;

	if( keys['D'] )										// Tilt Head Down When 'D' Is Pressed
		tilt += 1;

	if( tilt > 90.0 )									// Keep Tilt Between -90 and 90 Degrees
		tilt = 90.0;
	else if( tilt < -90.0 )
		tilt = -90.0;

	if( keys['W'] )										// Increase FOV When 'W' Is Pressed (Wide Angle)
	{
		fov++;
		glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
		glLoadIdentity();									// Reset The Projection Matrix
		gluPerspective(fov,(GLfloat)windowwidth/(GLfloat)windowheight,0.5f,50.0f);		// Recreate The Frustum
		glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	}
	else if( keys['T'] )								// Decrease FOV When 'T' Is Pressed (Telephoto or Zoom)
	{
		fov--;
		glMatrixMode(GL_PROJECTION);						// Select The Projection Matrix
		glLoadIdentity();									// Reset The Projection Matrix
		gluPerspective(fov,(GLfloat)windowwidth/(GLfloat)windowheight,0.5f,50.0f);		// Recreate The Frustum
		glMatrixMode(GL_MODELVIEW);							// Select The Modelview Matrix
	}

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer
	glLoadIdentity();									// Reset The Current Modelview Matrix

	glRotatef( tilt, 1.0, 0.0, 0.0 );					// Tilt Camera
	glRotatef( heading, 0.0, 1.0, 0.0 );				// Point Camera Towards Heading
	glTranslatef( xpos, -2.0, zpos );					// Translate To Camera Position

	glLightfv(GL_LIGHT1, GL_POSITION, LightPosition);	// Set Up Light Position

	ExtractFrustum();									// Extract The Frustum Plane Equations

	if( grid )
	{
		glPushMatrix();										// Save The Current MODELVIEW Matrix
		glColor3f( 0.6f, 0.6f, 0.6f );						// Draw Grid In GRAY
		glBegin( GL_LINES );								// Begin Drawing Lines For A Grid
		for( i = -100; i <= 100; i++ )						// Draw 200 Lines In One Direction
		{
			glVertex3i( i, 0, -100 );
			glVertex3i( i, 0, 100 );
		}
		for( i = -100; i <= 100; i++ )						// And 200 Lines Across The Others
		{
			glVertex3i( -100, 0, i );
			glVertex3i( 100, 0, i );
		}
		glEnd();
		glPopMatrix();										// Restore The Current MODELVIEW Matrix
	}

	for( i = 0; i < numobjects; i++ )						// Draw Up To 'numobjects' Objects
	{
		if( culling )										// Cull Objects If Culling Is On
		{
			if( mode == 0 )
				visible = PointInFrustum( objects[i][0], objects[i][1], objects[i][2] );
			else if( mode == 1 )
				visible = SphereInFrustum( objects[i][0], objects[i][1], objects[i][2], objects[i][3] );
			else
				visible = CubeInFrustum( objects[i][0], objects[i][1], objects[i][2], objects[i][3] );
		}
		else
			visible = true;									// If Not Culling, Objects Are Visible

		if( visible )
		{
			if( mode == 0 )														// Are We In Point Mode?
			{
				glColor3f( 1.0f, 1.0f, 1.0f );									// Draw Points In White
				glBegin( GL_POINTS );											// Begine Drawing Points
				glVertex3f( objects[i][0], objects[i][1], objects[i][2] );		// Draw The Point
				glEnd();														// End Drawing Points
				rendered++;														// Count How Many We Actually Draw
			}
			else if( mode == 1 )												// Are We In Sphere Mode?
			{
				glPushMatrix();													// Save The Current MODELVIEW Matrix
				glTranslatef( objects[i][0], objects[i][1], objects[i][2] );	// Translate To The Object's Location
				glColor3f( objects[i][4], objects[i][5], objects[i][6] );		// Set Drawing Color
				auxSolidSphere( objects[i][3] );								// Draw The Sphere
				glPopMatrix();													// Restore The Current MODELVIEW Matrix
				rendered++;														// Count How Many We Actually Draw
			}
			else																// Are We In Cube Mode?
			{
				glPushMatrix();													// Save The Current MODELVIEW Matrix
				glTranslatef( objects[i][0], objects[i][1], objects[i][2] );	// Translate To The Object's Location
				glColor3f( objects[i][4], objects[i][5], objects[i][6] );		// Set Drawing Color
				auxSolidCube( objects[i][3] * 2 );								// Draw The Cube
				glPopMatrix();													// Restore The Current MODELVIEW Matrix
				rendered++;														// Count How Many We Actually Draw
			}
		}
	}

	ortho();													// Enter Orthographic Mode
	glColor3f( 1.0f, 1.0f, 1.0f );								// Set Color To White
	glPrint( 1, 10, "FPS     : %0.1f", fps );					// Display Current FPS
	glPrint( 1, 20, "FOV     : %0.1f", fov  );					// Display Current FOV
	glPrint( 1, 30, "%s: %d", mode == 0 ? "Points  " : (mode == 1 ? "Spheres " : "Cubes   "), numobjects  );				// Display Total Number Of Objects
	glPrint( 1, 40, "Rendered: %d", rendered  );				// Display Number Of Objects Actually Drawn
	glPrint( 1, 50, "Culling : %s", culling ? "On" : "Off" );	// Display Culling State

	perspective();												// Return To Perspective Mode

	return true;												// Keep Going
}

GLvoid KillGLWindow(GLvoid)								// Properly Kill The Window
{
	if (fullscreen)										// Are We In Fullscreen Mode?
	{
		ChangeDisplaySettings(NULL,0);					// If So Switch Back To The Desktop
		ShowCursor(TRUE);								// Show Mouse Pointer
	}

	if (hRC)											// Do We Have A Rendering Context?
	{
		if (!wglMakeCurrent(NULL,NULL))					// Are We Able To Release The DC And RC Contexts?
		{
			MessageBox(NULL,"Release Of DC And RC Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}

		if (!wglDeleteContext(hRC))						// Are We Able To Delete The RC?
		{
			MessageBox(NULL,"Release Rendering Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}
		hRC=NULL;										// Set RC To NULL
	}

	if (hDC && !ReleaseDC(hWnd,hDC))					// Are We Able To Release The DC
	{
		MessageBox(NULL,"Release Device Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hDC=NULL;										// Set DC To NULL
	}

	if (hWnd && !DestroyWindow(hWnd))					// Are We Able To Destroy The Window?
	{
		MessageBox(NULL,"Could Not Release hWnd.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hWnd=NULL;										// Set hWnd To NULL
	}

	if (!UnregisterClass("OpenGL",hInstance))			// Are We Able To Unregister Class
	{
		MessageBox(NULL,"Could Not Unregister Class.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hInstance=NULL;									// Set hInstance To NULL
	}

	KillFont();										// Clean Up Font Data
}

/*	This Code Creates Our OpenGL Window.  Parameters Are:					*
 *	title			- Title To Appear At The Top Of The Window				*
 *	width			- Width Of The GL Window Or Fullscreen Mode				*
 *	height			- Height Of The GL Window Or Fullscreen Mode			*
 *	bits			- Number Of Bits To Use For Color (8/16/24/32)			*
 *	fullscreenflag	- Use Fullscreen Mode (TRUE) Or Windowed Mode (FALSE)	*/
 
BOOL CreateGLWindow(char* title, int width, int height, int bits, bool fullscreenflag)
{
	GLuint		PixelFormat;			// Holds The Results After Searching For A Match
	WNDCLASS	wc;						// Windows Class Structure
	DWORD		dwExStyle;				// Window Extended Style
	DWORD		dwStyle;				// Window Style
	RECT		WindowRect;				// Grabs Rectangle Upper Left / Lower Right Values
	WindowRect.left=(long)0;			// Set Left Value To 0
	WindowRect.right=(long)width;		// Set Right Value To Requested Width
	WindowRect.top=(long)0;				// Set Top Value To 0
	WindowRect.bottom=(long)height;		// Set Bottom Value To Requested Height

	fullscreen=fullscreenflag;			// Set The Global Fullscreen Flag

	hInstance			= GetModuleHandle(NULL);				// Grab An Instance For Our Window
	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraw On Size, And Own DC For Window.
	wc.lpfnWndProc		= (WNDPROC) WndProc;					// WndProc Handles Messages
	wc.cbClsExtra		= 0;									// No Extra Window Data
	wc.cbWndExtra		= 0;									// No Extra Window Data
	wc.hInstance		= hInstance;							// Set The Instance
	wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);			// Load The Default Icon
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
	wc.hbrBackground	= NULL;									// No Background Required For GL
	wc.lpszMenuName		= NULL;									// We Don't Want A Menu
	wc.lpszClassName	= "OpenGL";								// Set The Class Name

	if (!RegisterClass(&wc))									// Attempt To Register The Window Class
	{
		MessageBox(NULL,"Failed To Register The Window Class.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;											// Return FALSE
	}
	
	if (fullscreen)												// Attempt Fullscreen Mode?
	{
		DEVMODE dmScreenSettings;								// Device Mode
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));	// Makes Sure Memory's Cleared
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);		// Size Of The Devmode Structure
		dmScreenSettings.dmPelsWidth	= width;				// Selected Screen Width
		dmScreenSettings.dmPelsHeight	= height;				// Selected Screen Height
		dmScreenSettings.dmBitsPerPel	= bits;					// Selected Bits Per Pixel
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

		// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
		{
			// If The Mode Fails, Offer Two Options.  Quit Or Use Windowed Mode.
			if (MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","NeHe GL",MB_YESNO|MB_ICONEXCLAMATION)==IDYES)
			{
				fullscreen=FALSE;		// Windowed Mode Selected.  Fullscreen = FALSE
			}
			else
			{
				// Pop Up A Message Box Letting User Know The Program Is Closing.
				MessageBox(NULL,"Program Will Now Close.","ERROR",MB_OK|MB_ICONSTOP);
				return FALSE;									// Return FALSE
			}
		}
	}

	if (fullscreen)												// Are We Still In Fullscreen Mode?
	{
		dwExStyle=WS_EX_APPWINDOW;								// Window Extended Style
		dwStyle=WS_POPUP;										// Windows Style
		ShowCursor(FALSE);										// Hide Mouse Pointer
	}
	else
	{
		dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;			// Window Extended Style
		dwStyle=WS_OVERLAPPEDWINDOW;							// Windows Style
	}

	AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle);		// Adjust Window To True Requested Size

	// Create The Window
	if (!(hWnd=CreateWindowEx(	dwExStyle,							// Extended Style For The Window
								"OpenGL",							// Class Name
								title,								// Window Title
								dwStyle |							// Defined Window Style
								WS_CLIPSIBLINGS |					// Required Window Style
								WS_CLIPCHILDREN,					// Required Window Style
								0, 0,								// Window Position
								WindowRect.right-WindowRect.left,	// Calculate Window Width
								WindowRect.bottom-WindowRect.top,	// Calculate Window Height
								NULL,								// No Parent Window
								NULL,								// No Menu
								hInstance,							// Instance
								NULL)))								// Dont Pass Anything To WM_CREATE
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Window Creation Error.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	static	PIXELFORMATDESCRIPTOR pfd=				// pfd Tells Windows How We Want Things To Be
	{
		sizeof(PIXELFORMATDESCRIPTOR),				// Size Of This Pixel Format Descriptor
		1,											// Version Number
		PFD_DRAW_TO_WINDOW |						// Format Must Support Window
		PFD_SUPPORT_OPENGL |						// Format Must Support OpenGL
		PFD_DOUBLEBUFFER,							// Must Support Double Buffering
		PFD_TYPE_RGBA,								// Request An RGBA Format
		bits,										// Select Our Color Depth
		0, 0, 0, 0, 0, 0,							// Color Bits Ignored
		0,											// No Alpha Buffer
		0,											// Shift Bit Ignored
		0,											// No Accumulation Buffer
		0, 0, 0, 0,									// Accumulation Bits Ignored
		16,											// 16Bit Z-Buffer (Depth Buffer)  
		0,											// No Stencil Buffer
		0,											// No Auxiliary Buffer
		PFD_MAIN_PLANE,								// Main Drawing Layer
		0,											// Reserved
		0, 0, 0										// Layer Masks Ignored
	};
	
	if (!(hDC=GetDC(hWnd)))							// Did We Get A Device Context?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Create A GL Device Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if (!(PixelFormat=ChoosePixelFormat(hDC,&pfd)))	// Did Windows Find A Matching Pixel Format?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Find A Suitable PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if(!SetPixelFormat(hDC,PixelFormat,&pfd))		// Are We Able To Set The Pixel Format?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Set The PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if (!(hRC=wglCreateContext(hDC)))				// Are We Able To Get A Rendering Context?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Create A GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if(!wglMakeCurrent(hDC,hRC))					// Try To Activate The Rendering Context
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	ShowWindow(hWnd,SW_SHOW);						// Show The Window
	SetForegroundWindow(hWnd);						// Slightly Higher Priority
	SetFocus(hWnd);									// Sets Keyboard Focus To The Window
	ReSizeGLScene(width, height);					// Set Up Our Perspective GL Screen

	if (!InitGL())									// Initialize Our Newly Created GL Window
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Initialization Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	return TRUE;									// Success
}

LRESULT CALLBACK WndProc(	HWND	hWnd,			// Handle For This Window
							UINT	uMsg,			// Message For This Window
							WPARAM	wParam,			// Additional Message Information
							LPARAM	lParam)			// Additional Message Information
{
	switch (uMsg)									// Check For Windows Messages
	{
		case WM_ACTIVATE:							// Watch For Window Activate Message
		{
			if (!HIWORD(wParam))					// Check Minimization State
			{
				active=TRUE;						// Program Is Active
			}
			else
			{
				active=FALSE;						// Program Is No Longer Active
			}

			return 0;								// Return To The Message Loop
		}

		case WM_SYSCOMMAND:							// Intercept System Commands
		{
			switch (wParam)							// Check System Calls
			{
				case SC_SCREENSAVE:					// Screensaver Trying To Start?
				case SC_MONITORPOWER:				// Monitor Trying To Enter Powersave?
				return 0;							// Prevent From Happening
			}
			break;									// Exit
		}

		case WM_CLOSE:								// Did We Receive A Close Message?
		{
			PostQuitMessage(0);						// Send A Quit Message
			return 0;								// Jump Back
		}

		case WM_KEYDOWN:							// Is A Key Being Held Down?
		{
			keys[wParam] = TRUE;					// If So, Mark It As TRUE

			if( wParam == 'C' )						// Toggle Culling When 'C' Is Pressed
				culling = !culling;
			else if( wParam == 'G' )				// Toggle Grid Lines When 'G' Is Pressed
				grid = !grid;
			else if( wParam == 'M' )				// Change Mode
				mode = ++mode % 3;
			return 0;								// Jump Back
		}

		case WM_KEYUP:								// Has A Key Been Released?
		{
			keys[wParam] = FALSE;					// If So, Mark It As FALSE
			return 0;								// Jump Back
		}

		case WM_SIZE:								// Resize The OpenGL Window
		{
			ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));  // LoWord=Width, HiWord=Height
			return 0;								// Jump Back
		}
	}

	// Pass All Unhandled Messages To DefWindowProc
	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

int WINAPI WinMain(	HINSTANCE	hInstance,			// Instance
					HINSTANCE	hPrevInstance,		// Previous Instance
					LPSTR		lpCmdLine,			// Command Line Parameters
					int			nCmdShow)			// Window Show State
{
	MSG		msg;									// Windows Message Structure
	BOOL	done=FALSE;								// Bool Variable To Exit Loop

	// Ask The User Which Screen Mode They Prefer
	if (MessageBox(NULL,"Would You Like To Run In Fullscreen Mode?", "Start FullScreen?",MB_YESNO|MB_ICONQUESTION)==IDNO)
	{
		fullscreen=FALSE;							// Windowed Mode
	}

	// Create Our OpenGL Window
	if (!CreateGLWindow(TITLEBAR,640,480,16,fullscreen))
	{
		return 0;									// Quit If Window Was Not Created
	}

	while(!done)									// Loop That Runs While done=FALSE
	{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))	// Is There A Message Waiting?
		{
			if (msg.message==WM_QUIT)				// Have We Received A Quit Message?
			{
				done=TRUE;							// If So done=TRUE
			}
			else									// If Not, Deal With Window Messages
			{
				TranslateMessage(&msg);				// Translate The Message
				DispatchMessage(&msg);				// Dispatch The Message
			}
		}
		else										// If There Are No Messages
		{
			// Draw The Scene.  Watch For ESC Key And Quit Messages From DrawGLScene()
			if (active)								// Program Active?
			{
				if (keys[VK_ESCAPE])				// Was ESC Pressed?
				{
					done=TRUE;						// ESC Signalled A Quit
				}
				else								// Not Time To Quit, Update Screen
				{
					DrawGLScene();					// Draw The Scene
					SwapBuffers(hDC);				// Swap Buffers (Double Buffering)
				}
			}

			if (keys[VK_F1])						// Is F1 Being Pressed?
			{
				keys[VK_F1]=FALSE;					// If So Make Key FALSE
				KillGLWindow();						// Kill Our Current Window
				fullscreen=!fullscreen;				// Toggle Fullscreen / Windowed Mode
				// Recreate Our OpenGL Window
				if (!CreateGLWindow(TITLEBAR,640,480,16,fullscreen))
				{
					return 0;						// Quit If Window Was Not Created
				}
			}
		}
	}

	// Shutdown
	KillGLWindow();									// Kill The Window
	return (msg.wParam);							// Exit The Program
}
